# -*- coding: utf-8; tab-width: 4; indent-tabs-mode: nil -*-
#
# Copyright (C) 2009 - Jeremy AustinBardo <tjaustinbardo@gmail.com>;
#   Adam J. Lincoln <adamjlincoln@gmail.com>; Richard Harding <rharding@raken>;
#   Neil Shepperd <nshepperd@gmail.com>
# Copyright (C) 2006 - Florian Ludwig <dino@phidev.info>;
#   Franz Pletz <fpletz@franz-pletz.org>
#
# Apt-zeroconf is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Apt-zeroconf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with apt-zeroconf; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""
Package Monitoring Service Module for apt-zeroconf. Watch the apt cache for new
or removed packages to keep a list of locally available packages.

@requires: pyinotify.[WatchManager|ThreadedNotifier|ProcessEvent]
@requires: logging
@requires: threading
@requires: L{aptzeroconf.util.synchronized}
@requires: L{aptzeroconf.config.config}
@requires: L{aptzeroconf.avaphi.hosts}
"""
from pyinotify import WatchManager, ThreadedNotifier, ProcessEvent
import logging, threading
from aptzeroconf.util import synchronized
from aptzeroconf.config import config

CACHE_PATH = config['apt-cache-path']
LIST_SEP = '\n'

def _get_eventmask():
    """
    Set ProcessEvents to watch for with pyinotify.

    @requires: pyinotify
    @note: Newer pyinotify versions have dropped child methods with children. So
        this helps ensure AZC works for prior versions as well as future ones.
    """
    import pyinotify

    notify = (pyinotify if hasattr(pyinotify,'IN_MOVED_TO')
        else pyinotify.EventsCodes)

    return (notify.IN_MOVED_TO | notify.IN_DELETE | notify.IN_OPEN |
            notify.IN_CLOSE_WRITE)


class _AptWatcherProcess(ProcessEvent):
    """
    Process events triggered by pyinotify.watchmanager.

    @requires: threading
    @requires: pyinotify.ProcessEvent
    @requires: aptzeroconf.util.synchronized
    @cvar _newpkgcallback: Instances of package added callback method set by
        L{PkgList} instance.
    @cvar _delpkgcallback: Instances of package removed callback method set by
        L{PkgList} instance.
    @cvar _cachelockcallback: Instances of cache lock callback method set by
        L{PkgList} instance.
    @cvar _cacheunlockcallback: Instances of cache unlock callback method set by
        L{PkgList} instance.
    @cvar lock: Thread lock.
    """
    _newpkgcallback = []
    _delpkgcallback = []
    _cachelockcallback = []
    _cacheunlockcallback = []
    lock = threading.Lock()

    @synchronized(lock)
    def process_IN_MOVED_TO(self, event):
        for func in self.NewPkgCallback:
            func(event.name)

    @synchronized(lock)
    def process_IN_DELETE(self, event):
        for func in self.DelPkgCallback:
            func(event.name)

    @synchronized(lock)
    def process_IN_OPEN(self, event):
        if event.name == 'lock':
            for func in self.CacheLockCallback:
                func()

    @synchronized(lock)
    def process_IN_CLOSE_WRITE(self, event):
        if event.name == 'lock':
            for func in self.CacheUnlockCallback:
                func()

    @property
    def NewPkgCallback(self):
        logging.debug('[APT WATCH] package addition notification.')
        return self._newpkgcallback

    @NewPkgCallback.setter
    def NewPkgCallback(self, func):
        self._newpkgcallback.append(func)

    @property
    def DelPkgCallback(self):
        logging.debug('[APT WATCH] package deletion notification.')
        return self._delpkgcallback

    @DelPkgCallback.setter
    def DelPkgCallback(self, func):
        self._delpkgcallback.append(func)

    @property
    def CacheLockCallback(self):
        logging.debug('[APT WATCH] package access notification.')
        return self._cachelockcallback

    @CacheLockCallback.setter
    def CacheLockCallback(self, func):
        self._cachelockcallback.append(func)

    @property
    def CacheUnlockCallback(self):
        logging.debug('[APT WATCH] package release notification.')
        return self._cacheunlockcallback

    @CacheUnlockCallback.setter
    def CacheUnlockCallback(self, func):
        self._cacheunlockcallback.append(func)

class PkgList():
    """
    Create an AZC cache service cache package list using apt_watcher \
    pyinotifier instance.

    @requires: threading
    @requires: aptzeroconf.util.synchronized
    """
    list_lock = threading.Lock()
    queues_lock = threading.Lock()
    urllist = {}

    def __init__(self, notifycallback):
        """
        Initialiaze an AZC cache service cache package list. Set callbacks for
        new, deleted and unlock pyinotify events.

        @param notifycallback: apt_watcher instance
        @type notifycallback: object
        @requires: os.listdir
        """
        import os

        self.pkglist = set(filter(lambda s: s.endswith('.deb'),
            os.listdir(CACHE_PATH)))

        self.add_queue = []
        self.del_queue = []

        notifycallback.NewPkgCallback = lambda p: self.add_queue.append(p)
        notifycallback.DelPkgCallback = lambda p: self.del_queue.append(p)
        notifycallback.CacheUnLockCallback = self._unlocked

    def __str__(self):
        """
        Retun a listing of packages in local cache.

        @return: Packages in local cache
        @rtype: basestring
        """
        return LIST_SEP.join(self.pkglist)

    @synchronized(queues_lock)
    def _unlocked(self):
        """
        Unlock package callback to finalize new package list after addition or
        removal of a package.

        @requires: aptzeroconf.avaphi.hosts
        """
        from aptzeroconf.avaphi import hosts
        self.list_lock.acquire()
        if self.add_queue:

            self.pkglist |= self.add_queue
            logging.debug('[APT LIST] adding packages.')
            self.add_queue = []
        if self.del_queue:

            self.pkglist -= self.del_queue
            logging.debug('[APT LIST] removing packages.')
            self.del_queue = []
        self.list_lock.release()

        hosts.propagate_update()
        logging.debug('[APT LIST] hosts\' update request made.')

    @synchronized(list_lock)
    def has(self, name):
        """
        Inquire with local AZC cache service if package is in the list.

        @keyword name: filename
        @type name: basestring
        @return: name is in pkglist
        @rtype: boolean
        """
        return name in self.pkglist

    def get_path(self, name):
        """
        Get packages full local url in local AZC cache service.

        @param name: Filename
        @type name: basestring
        @return: Full path for name
        @rtype: basestring
        @requires: os.path.join
        """
        from os.path import join
        return join(CACHE_PATH, name)

_watchmanager = WatchManager()  #: Watch Manager
_apt_watcher_process = _AptWatcherProcess() #: Watch Process
_watchmanager.add_watch(CACHE_PATH, _get_eventmask()) #: Watch Events
apt_watcher = ThreadedNotifier(_watchmanager, _apt_watcher_process)
pkglist = PkgList(_apt_watcher_process)
